/*
 *	An example from the book "Start Programming Using HTML, CSS, and JavaScript" by Iztok Fajfar.
 *	Copyright (c) 2015, by Taylor and Francis Group, LLC
 */
var Cell = function() { 
	this.number = null;
	this.type = "answer";
};	

Cell.prototype.setNumber = function(num) {
	this.number = num;
};

Cell.prototype.getNumber = function() {
	return this.number;
};

Cell.prototype.getType = function() {
	return this.type;
};		

Cell.prototype.setType = function(type) {
//type: "clue", "answer", "wrong", "candidate"
	this.type = type;
};	

var Sudoku = function() {
	var i, j;
	
	this.puzzle = null;
	this.board = [];
	
	for (i = 0; i < 9; i++) {
		this.board[i] = [];
		for (j = 0; j < 9; j++) {
			this.board[i][j] = new Cell();
		}
	}	
};		

Sudoku.prototype.setNumberManually = function(number, row, col, type) {
	this.board[row][col].setNumber(number);
	this.board[row][col].setType(type);
}; 

Sudoku.prototype.setClues = function(initial) {
	for (i = 0; i < 9; i++) {
		for (j = 0; j < 9; j++) {
			this.board[i][j].setType("answer");
			this.board[i][j].setNumber(initial[i][j]);
			if (initial[i][j] !== null) {
				this.board[i][j].setType("clue");
			}
		}
	}
};

Sudoku.prototype.setNumber = function(number, row, col) {
	var i, j;
	var iStart, jStart;

	//Removes the current number
	this.board[row][col].setNumber(null);
	
	//Checks a row
	for (i = 0; i < 9; i++) {
		if (this.board[row][i].getNumber() == number) {
			return false;
		}
	}

	//Checks a column
	for (i = 0; i < 9; i++) {
		if (this.board[i][col].getNumber() == number) {
			return false;
		}
	}

	//Checks a 3x3 box
	iStart = row - row % 3;
	jStart = col - col % 3;
	for (i = iStart; i < iStart + 3; i++) {
		for (j = jStart; j < jStart + 3; j++) {
			if (this.board[i][j].getNumber() == number) {
				return false;
			}
		}
	}
	this.board[row][col].setNumber(number);
	return true;
};			
 
Sudoku.prototype.start = function(id) {
	var row, cell, text, number;
	var puzzleContainer;
	var newButton, form;
		
	puzzleContainer = document.getElementById(id);
	form = document.createElement("form");
	this.puzzle = document.createElement("table");
	this.puzzle.className = "sudoku-table";
	puzzleContainer.appendChild(form);
	newButton = document.createElement("input");
	newButton.type = "button";
	newButton.value = "New Puzzle";
	newButton.className = "sudoku-button";
	newButton.onclick = insertNumbers(this, false);
	
	form.appendChild(this.puzzle);
	form.appendChild(newButton);
	
	for (i = 0; i < 9; i++) {
		row = document.createElement("tr");
		this.puzzle.appendChild(row);
		for (j = 0; j < 9; j++) {
			cell = document.createElement("td");
			row.appendChild(cell);
			text = document.createElement("input");
			text.type = "text";
			text.className = "answer";
			cell.appendChild(text);
			text.oninput = handleInput(this, i, j);
		}
	}

	insertNumbers(this, true)();
};

var insertNumbers = function (s, fromLocalStorage) {
	return function() {
		var i, j;
		var text, number;
		var inputs = s.puzzle.getElementsByTagName("input");
		var g;
		if (!fromLocalStorage) {
			MyStorage.clear();
		}
		if (!MyStorage.isAvailable()) {
			g = new SudokuGenerator();
			do { //Repeats until we have a single-solution Sudoku.
				g.reset(); //Loads a completely solved puzzle.
				g.shuffle(); //Swaps rows and columns randomly.
				g.removeNumbers(30); //Leaves 30 clues in the puzzle.
				s.setClues(g.getClues()); //Initializes the Sudoku using the obtained clues.
			} while (!s.isSolutionUnique()); //Checks whether this is a single-solution Sudoku.
		}
		else {	//Read from localStorage:
			for (i = 0; i < 9; i++) {
				for (j = 0; j < 9; j++) {
					s.setNumberManually(
						MyStorage.getValue("value", i, j), i, j,  
						MyStorage.getValue("class", i, j)
					);
				}
			}
		}
		for (i = 0; i < 9; i++) {
			for (j = 0; j < 9; j++) {
				text = inputs[i * 9 + j];
				if (s.board[i][j].getType() == "clue") {
					text.value = s.board[i][j].getNumber();
					text.readOnly = true;
					text.className = "clue";
				}
				else {
					text.value = s.board[i][j].getNumber();
					if (text.value === null) {
						text.value = "";
					}
					text.className = s.board[i][j].getType();
					text.readOnly = false;
				}
			}
		}
	};
};

var handleInput = function(sdk, row, col) {
	return function() {
		var v = parseInt(this.value);
		this.className = "answer";
		if (v >= 1 && v <= 9) {
			if (!sdk.setNumber(this.value, row, col)) {
				this.className = "wrong";
			}
		}
		else {
			sdk.setNumber(null, row, col);
			this.className = "candidate";
		}
	};
};

/*
 * Constructs a SudokuGenerator object instance. After constructing
 * an object instance, you should call the methods reset(),
 * shuffle(), removeNumbers(), and getClues().
 *
 * Example:
 *   var clues, g = new SudokuGenerator();
 *   g.reset();
 *   g.shuffle();
 *   g.removeNumbers(30);
 *   clues = g.getClues();
 */
var SudokuGenerator = function() {
	this.clues = null;
};
/*
 * Resets a SudokuGenerator object. Whenever you want to re-create
 * a puzzle, this method should be called first.
 */
SudokuGenerator.prototype.reset = function() {
	this.clues = [
		[8, 2, 7, 1, 5, 4, 3, 9, 6],
		[9, 6, 5, 3, 2, 7, 1, 4, 8],
		[3, 4, 1, 6, 8, 9, 7, 5, 2],
		[5, 9, 3, 4, 6, 8, 2, 7, 1],
		[4, 7, 2, 5, 1, 3, 6, 8, 9],
		[6, 1, 8, 9, 7, 2, 4, 3, 5],
		[7, 8, 6, 2, 3, 5, 9, 1, 4],
		[1, 5, 4, 7, 9, 6, 8, 2, 3],
		[2, 3, 9, 8, 4, 1, 5, 6, 7],
	];
};

/*
 * Shuffles the numbers in the puzzle grid.
 */
SudokuGenerator.prototype.shuffle = function() {
	var i;
	for (i = 0; i < 500; i++) {
		this.swapBlockCols(rand(3), rand(3));
		this.swapBlockRows(rand(3), rand(3));
		this.swapColsInBlock(rand(3), rand(3), rand(3));
		this.swapRowsInBlock(rand(3), rand(3), rand(3));
	}
};

/*
 * Swaps two rows.
 * Parameters:
 *   r1, r2: the indexes of rows to be swapped (indexes between zero and eight)
 */
SudokuGenerator.prototype.swapRows = function(r1, r2) {
	var i, tmp;
	for (i = 0; i < 9; i++) {
		tmp = this.clues[r1][i];
		this.clues[r1][i] = this.clues[r2][i];
		this.clues[r2][i] = tmp;
	}
};

/*
 * Swaps two columns.
 * Parameters:
 *   c1, c2: the indexes of columns to be swapped (indexes between zero and eight)
 */
SudokuGenerator.prototype.swapCols = function(c1, c2) {
	var i, tmp;
	for (i = 0; i < 9; i++) {
		tmp = this.clues[i][c1];
		this.clues[i][c1] = this.clues[i][c2];
		this.clues[i][c2] = tmp;
	}
};

/*
 * Swaps two rows within a row of 3x3 blocks.
 * Parameters:
 *   br: the index of a row of 3x3 blocks (an integer between zero and two)
 *   r1, r2: the indexes of rows to be swapped. Indexes are
 *           relative to the given row of 3x3 blocks and can be
 *           integers between zero and two.
 */
SudokuGenerator.prototype.swapRowsInBlock = function(br, r1, r2) {
	this.swapRows(br * 3 + r1, br * 3 + r2);
};

/*
 * Swaps two columns within a column of 3x3 blocks.
 * Parameters:
 *   bc: the index of a column of 3x3 blocks (an integer between zero and two)
 *   c1, c2: the indexes of columns to be swapped. Indexes are
 *           relative to the given column of 3x3 blocks and can be
 *           integers between zero and two.
 */
SudokuGenerator.prototype.swapColsInBlock = function(bc, c1, c2) {
	this.swapCols(bc * 3 + c1, bc * 3 + c2);
};

/*
 * Swaps two rows of 3x3 blocks.
 * Parameters:
 *   r1, r2: the indexes of rows to be swapped (integers between zero and two)
 */
SudokuGenerator.prototype.swapBlockRows = function(r1, r2) {
	for (i = 0; i < 3; i++) {
		this.swapRows(r1 * 3 + i, r2 * 3 + i);
	}
};

/*
 * Swaps two columns of 3x3 blocks.
 * Parameters:
 *   c1, c2: the indexes of columns to be swapped (integers between zero and two)
 */
SudokuGenerator.prototype.swapBlockCols = function(c1, c2) {
	for (i = 0; i < 3; i++) {
		this.swapCols(c1 * 3 + i, c2 * 3 + i);
	}
};

/*
 * Randomly removes numbers from the puzzle grid by setting them to null.
 * Parameters:
 *   numClues: the number of clues to be left in the grid
 */
SudokuGenerator.prototype.removeNumbers = function(numClues) {
	var i, j;
	var remain = 81;
	while (remain > numClues) {
		i = rand(9);
		j = rand(9);
		if (this.clues[i][j] != null) {
			this.clues[i][j] = null;
			remain--;
		}
	}
};

/*
 * Returns a two-dimensional array of clues. Empty cells are set to null.
 */
SudokuGenerator.prototype.getClues = function() {
	return this.clues;
};

/*
 * Returns a non-negative random integer.
 * Parameters:
 *   n: the upper limit of a returned number
 * Returns:
 *   A random integer between zero and n - 1
 */
var rand = function(n) {
	return Math.floor(Math.random() * n);
};

/*
 * Checks whether a puzzle has a unique solution.
 * Returns:
 *   If the puzzle has a single solution, then the method returns true.
 */
Sudoku.prototype.isSolutionUnique = function() {
	var i, j;
	this.empty = [];
	this.solutions = 0;
	//Search for all the empty cells:
	for (i = 0; i < 9; i++) {
		for (j = 0; j < 9; j++) {
			if (this.board[i][j].getNumber() == null) {
			//Store the row and column indexes of the cell into the array empty:
				this.empty.push({row: i, col: j});
			}
		}
	}
	this.checkCell(0);
	//The checkCell() recursive function stores the number of
	//solutions in the property solutions. If there is more than one
	//solution, then solutions is set to two.
	if (this.solutions == 1) {
		return true;
	}
	else {
		return false;
	}
};

/*
 * This is a recursive function used by the isSolutionUnique() method.
 */
Sudoku.prototype.checkCell = function(n) {
	var number;
	for (number = 1; number < 10; number++) {
	//Try to insert numbers from one to nine, one at a time.
		if (this.setNumber(number, this.empty[n].row, this.empty[n].col)) {
		//If the insertion was successful.
			if (n == this.empty.length - 1) {
			//If the last of the empty cells was filled, we have a solution.
				this.solutions++;
			}
			else {
			//Otherwise, go and check the next of the empty cells recursively.
				this.checkCell(n + 1);
			}
		}
		if (this.solutions > 1) {
		//More than one solution already found: no need to continue.
			return;
		}
	}
	//All the numbers have been tried for this cell, so clear the
	//cell before returning:
	this.setNumber(null, this.empty[n].row, this.empty[n].col);
};

var MyStorage = function() {
};

MyStorage.setValue = function(item, val, row, col) {
	localStorage.setItem(item + row + col, val);
};

MyStorage.getValue = function(item, row, col) {
	var val = localStorage.getItem(item + row + col);
	if (item == "value") {
		val = Number(val);
		if (val !== val || val == 0) {
			val = null;
		}
	}
	return val;
};

MyStorage.isAvailable = function() {
	return Boolean(localStorage.length);
};

MyStorage.clear = function() {
	localStorage.clear();
};

window.onload = function() {
	(new Sudoku()).start("sudoku-puzzle");
};

window.onunload = function() {
	var i, j, inputs = document.getElementById("sudoku-puzzle").getElementsByTagName("input");
	MyStorage.clear();
	for (i = 0; i < 9; i++) {
		for (j = 0; j < 9; j++) {	
			MyStorage.setValue("value", inputs[i * 9 + j].value, i, j);
			MyStorage.setValue("class", inputs[i * 9 + j].className, i, j);
		}
	}
};
